/**
 * FOSS Galaxy Game Library
 * Copyright (c) 2022 Joseph Walton-Rivers
 * webpigeon@fossgalaxy.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef FGGL_APP_H
#define FGGL_APP_H

#define assertm(exp, msg) assert(((void)msg, exp))

#include <cassert>
#include <string>
#include <memory>
#include <unordered_map>

#include <fggl/ecs3/types.hpp>
#include <fggl/ecs3/module/module.h>
#include <fggl/gfx/window.hpp>
#include <fggl/gfx/paint.hpp>
#include <fggl/util/states.hpp>

namespace fggl {

    class App;
    class AppState;

    using Identifer = std::string;
    using AppMachine = util::StateMachine<AppState, Identifer>;

    class AppState {
        public:
            /**
             * Create an Application State.
             *
             * A state is responsible for managing user interaction with the app. When created, the appstate
             * is passed a reference to the application that owns it. The lifetime of the state is bounded
             * by the lifetype of this object.
             *
             * @param owner a non-owned reference to the owner of the state.
             */
            explicit AppState(App& owner) : m_owner( owner ) {}

            /**
             * Update the underlying model of this state.
             *
             * States should not assume that one update means one render call, as the game loop may issue
             * multiple updates per render or vice-versa depending on requriements. Update is intended for
             * dispatching game-system related infomation.
             */
            virtual void update() = 0;

            /**
             * Perform actions neccerary for rendering the scene.
             * 
             * When this method is invoked it is safe to assume that rendering of some form will take place
             * after it returns and the rendering state should be updated to reflect this. The rendering
             * environment will be passed in as an argument.
             *
             * It is not safe to assume the render target will always be the same, as the scene may be
             * rendered in mutliple passes (eg, for VR requirements).
             */
            virtual void render(gfx::Paint& paint) = 0;

            virtual void activate() {}
            virtual void deactivate() {}

        protected:
            App& m_owner;
    };

    class App {
        public:
            explicit App(const Identifer& name);
            App(const Identifer& name, const Identifer& folderName);

            // class is non copy-able
            App(const App& app) = delete;
            App& operator=(App other) = delete;

            inline void setWindow( std::unique_ptr<gfx::Window>&& window ) {
                m_window = std::move(window);
            }

            /**
             * Perform main game loop functions.
             */
            int run(int argc, const char** argv);

            template<typename T>
            T& add_state(const Identifer& name) {
                static_assert( std::is_base_of<AppState,T>::value, "States must be AppStates");
                return m_states.put<T>(name, *this);
            }

            template<typename T, typename... Args>
            T& use(Args&& ...args) {
                auto ptr = m_modules->load<T>(args...);
                return *ptr;
            }

            inline void change_state(const Identifer& name) {
                m_states.active().deactivate();
                m_states.change(name);
                m_states.active().activate();
            }

            inline AppState& active_state() const {
                return m_states.active();
            }

            inline bool running() const {
                return m_running;
            }

            inline void running(bool state) {
                m_running = state;
            }

        private:
            bool m_running;
            std::unique_ptr<ecs3::TypeRegistry> m_types;
            std::unique_ptr<ecs3::ModuleManager> m_modules;
            std::unique_ptr<gfx::Window> m_window;
            AppMachine m_states;
    };

}

#endif
