/**
 * FOSS Galaxy Game Library
 * Copyright (c) 2022 Joseph Walton-Rivers
 * webpigeon@fossgalaxy.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef FGGL_UTIL_STATES_H
#define FGGL_UTIL_STATES_H

#include <cassert>
#include <string>
#include <memory>
#include <unordered_map>

namespace fggl::util {

    template<typename S, typename I>
    class StateMachine {
        public:
            using Identifer = I;
            using StateType = S;

            StateMachine() = default;
            ~StateMachine() = default;

            // class is non copy-able
            StateMachine(const StateMachine& app) = delete;
            StateMachine& operator=(StateMachine other) = delete;

            template<typename T, typename... Args>
            T& put(const Identifer& name, Args&&... args) {
                static_assert( std::is_base_of<StateType,T>::value, "States must be AppStates");
                m_states[name] = std::make_unique<T>( std::forward<Args...>(args...) );

                // no active scene? first is the active one
                if ( m_active.empty() ) {
                    m_active = name;
                }

                return *(T*)(m_states[name].get());
            }

            void change(const Identifer& name) {
                assertm( m_states.find(name) != m_states.end(), "state does not exist");

                active().deactivate();
                m_active = name;
                active().activate();
            }

            StateType& active() const {
                assertm( m_states.find(m_active) != m_states.end(), "active state does not exist!");
                return *(m_states.at( m_active ).get());
            }

        private:
            Identifer m_active;
            std::unordered_map<Identifer, std::unique_ptr<StateType>> m_states;

    };

}

#endif
